package com.formos.tapestry.xpath;

import java.util.ArrayList;
import java.util.List;

import org.apache.tapestry5.dom.Attribute;
import org.apache.tapestry5.dom.Element;
import org.apache.tapestry5.dom.Node;
import org.jaxen.BaseXPath;
import org.jaxen.Context;
import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.function.StringFunction;

/**
 * Implements XPath for the Tapestry DOM.
 *
 * @author Paul Field
 */
@SuppressWarnings("unchecked") // Jaxen API does not use generics
public class TapestryXPath extends BaseXPath implements XPath {
    private static final long serialVersionUID = 7040266295703699568L;
    private static final TapestryXPath TEXT_DESCENDANTS;
    static {
        try {
            TEXT_DESCENDANTS = xpath("descendant::text()");
        } catch (JaxenException e) {
            throw new TapestryXPathException(e);
        }
    }

    public TapestryXPath(String xpath) throws JaxenException {
        super(xpath, TapestryNavigator.TAPESTRY_NAVIGATOR);
    }


    /**
     * Creates a new TapestryXPath object.
     *
     * @param xpath the XPath expression as a string (not null)
     * @return the new instance (not null)
     */
    public static TapestryXPath xpath(String xpath) throws JaxenException {
        return new TapestryXPath(xpath);
    }


    @Override
    public List<Node> selectNodes(Object node) throws JaxenException {
    	return super.selectNodes(node);
    }


    /**
     * As {@link #selectNodes(Object)} but assumes the result will be a collection of
     * Elements.
     */
    public List<Element> selectElements(Object node) throws JaxenException {
        return super.selectNodes(node);
    }


    /**
     * As {@link #selectNodes(Object)} but assumes the result will be a collection of
     * Attributes.
     */
	public List<Attribute> selectAttributes(Object node) throws JaxenException {
        return super.selectNodes(node);
	}



    /**
     * As {@link #selectNodes(Object)} but converts each of the selected nodes into its XPath string-value.
     * @see <a href="http://www.w3.org/TR/xpath#dt-string-value">XPath Spec: string-value</a>
     * @see #normalizedDescendantText(Node)
     */
    public List<String> stringValuesOf(Object node) throws JaxenException {
        Context context = getContext(node);
        List<Object> resultNodes = selectNodesForContext(context);
        List<String> resultStrings = new ArrayList<String>(resultNodes.size());
        for (Object result : resultNodes) {
            resultStrings.add(StringFunction.evaluate(result, context.getNavigator()));
        }

        return resultStrings;
    }


    /**
     * Returns the {@link Element#getChildMarkup()} from each element identified by {@link #selectElements(Object)}.
     */
    public List<String> selectElementsChildMarkup(Object node) throws JaxenException {
        List<Element> selectedNodes = selectElements(node);
        List<String> childMarkup = new ArrayList<String>(selectedNodes.size());
        for (Element element : selectedNodes) {
            childMarkup.add(element.getChildMarkup());
        }
        return childMarkup;
    }


    /**
     * Returns the {@link Element#getAttribute(String)} from each element identified by {@link #selectElements(Object)}.
     * Note this is different from using XPath to match the attributes - if the attribute is missing then XPath can't match it
     * whereas this will include a "null" in the list.
     */
    public List<String> selectElementsAttribute(Object node, String attributeName) throws JaxenException {
        List<Element> selectedNodes = selectElements(node);
        List<String> attributeValues = new ArrayList<String>(selectedNodes.size());
        for (Element element : selectedNodes) {
            attributeValues.add(element.getAttribute(attributeName));
        }
        return attributeValues;
    }


    /**
     * As {@link #selectSingleNode(Object)} but assumes the result will be an Element.
     */
    public Element selectSingleElement(Object node) throws JaxenException {
        return (Element)selectSingleNode(node);
    }


	/**
     * Selects a single node and then concatenates all the descendant text nodes and
     * applies {@link #normalizeText(String)}. Useful to get a sense of what would be
     * displayed as text to a user by a browser.
     *
     * @see #selectSingleNode(Object)
     * @throws JaxenException
     */
	public String singleNormalizedDescendantText(Object node) throws JaxenException {
        Object foundNode = selectSingleNode(node);
        List<String> textNodeValues = TEXT_DESCENDANTS.stringValuesOf(foundNode);
        return normalizeText(concatenate(textNodeValues));
	}


	/**
     * Selects nodes and then, for each node found, concatenates all the descendant
     * text nodes and applies {@link #normalizeText(String)}. Useful to get a sense of what would be
     * displayed as text to a user by a browser.
     *
     * @return the list of text (not null; same size as the number of nodes selected by the XPath)
     * @see #selectSingleNode(Object)
     * @throws JaxenException
     */
	public List<String> normalizedDescendantText(Object node) throws JaxenException {
        List<Node> nodes = selectNodes(node);
        List<String> results = new ArrayList<String>(nodes.size());
        for (Node foundNode : nodes) {
            List<String> textNodeValues = TEXT_DESCENDANTS.stringValuesOf(foundNode);
            results.add(normalizeText(concatenate(textNodeValues)));
        }
        return results;
    }


    /**
     * Normalises text from the DOM so that is more like what is displayed to a user by a browser
     * (for Latin scripts):
     * whitespace is replaced with space, multiple space is collapsed into single spaces and any space at
     * the beginning or end is trimmed.
     */
    public static String normalizeText(String text) {
        StringBuilder result = new StringBuilder(text.length());
        boolean skippingSpaces = true;
        for (int i=0; i < text.length(); i++) {
            char c = text.charAt(i);
            if (Character.isWhitespace(c) || c == 160) {
                if (!skippingSpaces) {
                    result.append(' ');
                }
                skippingSpaces = true;
            } else {
                result.append(c);
                skippingSpaces = false;
            }
        }
        if (skippingSpaces == true && result.length() > 0) {
            result.deleteCharAt(result.length()-1);
        }

        return result.toString();
    }


    private static String concatenate(List<String> strings) {
        if (strings.isEmpty()) {
            return "";
        } else if (strings.size() == 1) {
            return strings.get(0);
        } else {
            int maxLength = 0;
            for (String string : strings) {
                maxLength += string.length();
            }
            StringBuilder concatenated = new StringBuilder(maxLength);
            for (String string : strings) {
                concatenated.append(string);
            }
            return concatenated.toString();
        }
    }

}